﻿
using System.Diagnostics;

record Corn {
	public int Weight { get; init; }
	public float Quality { get; init; }
}

class CropFarmer {
	public CropFarmer(float cornQuality, int initialWeightInStock) {
		_cornQuality = cornQuality;
		_weightInStock = initialWeightInStock;
	}

	private float _cornQuality;
	private int _weightInStock;

	public Task<Corn>? OrderCorn(int targetWeight, CancellationToken ct) {
		Console.WriteLine($"### CropFarmer: got order for {targetWeight} kg of corn ...");
		if (ct.IsCancellationRequested) {
			Console.WriteLine("### CropFarmer: order cancellation requested -> CANCELLING ORDER !!!");
			return Task.FromCanceled<Corn>(ct);
		}
		if (targetWeight <= _weightInStock) {
			Console.WriteLine($"### CropFarmer: SHIPPING {targetWeight} kg of corn ~~ FROM STOCK ~~ ...");
			_weightInStock -= targetWeight;
			return Task.FromResult(new Corn { Weight = targetWeight, Quality = _cornQuality });
		} else {
			Console.WriteLine($"### CropFarmer: {targetWeight} kg of corn not available, cancelling corn order ...");
			return null;
		}
	}
}

record Pork {
	public float Weight { get; init; }
}

class PigProducer {
	private const int CornWeightPerPig = 50;
	private const float PorkPerPig = 0.2f;

	public required CropFarmer FavoriteFarmer { get; init; }

	public int GetPigsNeededForPork(float targetPorkWeight) => (int) Math.Round(targetPorkWeight / PorkPerPig, MidpointRounding.ToPositiveInfinity);

	public Task<Pork>? OrderPork(int pigCount, CancellationToken ct) {
		Console.WriteLine($"++ PigProducer: got order for {pigCount} pigs worth of pork ...");
		if (ct.IsCancellationRequested) {
			Console.WriteLine("++ PigProducer: order cancellation requested -> CANCELLING ORDER !!!");
			return Task.FromCanceled<Pork>(ct);
		}

		int cornNeeded = pigCount * CornWeightPerPig;

		Task<Corn>? futureCorn = FavoriteFarmer.OrderCorn(cornNeeded, ct);
		if (futureCorn is null) {
			Console.WriteLine($"++ PigProducer: unable to order {cornNeeded} kg of corn, cancelling pork order ...");
			return null;
		}
		Console.WriteLine($"++ PigProducer: corn order successful, order state: futureCorn.IsCompleted -> {futureCorn.IsCompleted} ...");

		TaskCompletionSource<Pork> porkPromise = new TaskCompletionSource<Pork>();
		ThreadPool.QueueUserWorkItem(_ => {
			Console.WriteLine("++ PigProducer: building pig pen ...");
			Thread.Sleep(500);

			Console.WriteLine($"++ PigProducer: pen is ready, need to feed pigs ... Is futureCorn ready? .IsCompleted -> {futureCorn.IsCompleted}");
			Corn corn = futureCorn.Result;
			Console.WriteLine($"++ PigProducer: got {corn.Weight} kg of corn of {(int) (corn.Quality * 100)} % quality, feeding to pigs !!! TO USE FIRST CORN !!!");
			for (int i = 0; i < 10; i++) {
				if (ct.IsCancellationRequested) {
					Console.WriteLine("++ PigProducer: order cancellation requested -> CANCELLING ORDER !!!");
					porkPromise.SetCanceled(ct);
					return;
				}
				Thread.Sleep(150);
			}
			int pigsProduced = (int) (corn.Weight * corn.Quality / CornWeightPerPig);
			Console.WriteLine($"++ PigProducer: {pigsProduced} pigs produced ...");

			var producedPork = new Pork { Weight = pigsProduced * PorkPerPig };
			Console.WriteLine($"++ PigProducer: SHIPPING {producedPork.Weight} kg of pork from {pigCount} pigs >> VIA PROMISE >> ...");
			porkPromise.SetResult(producedPork);
		});

		// Return future pork:
		return porkPromise.Task;
	}
}

record HamburgerPatty { }

class MeatProcessingPlant {
	private const float PorkPerHamburgerPetty = 0.1f;

	public required PigProducer FavoritePigProducer { get; init; }

	public Task<IReadOnlyList<HamburgerPatty>>? OrderHamburgerPatties(int count, CancellationToken ct) {
		Console.WriteLine($"> MeatProcessingPlant: got order for {count} hamburger petties ...");
		if (ct.IsCancellationRequested) {
			Console.WriteLine("> MeatProcessingPlant: order cancellation requested -> CANCELLING ORDER !!!");
			return Task.FromCanceled<IReadOnlyList<HamburgerPatty>>(ct);
		}

		float porkNeeded = count * PorkPerHamburgerPetty;

		int pigsNeeded = FavoritePigProducer.GetPigsNeededForPork(porkNeeded);

		Task<Pork>? futurePork = FavoritePigProducer.OrderPork(pigsNeeded, ct);
		if (futurePork is null) {
			Console.WriteLine($"> MeatProcessingPlant: unable to order {pigsNeeded} pigs for {porkNeeded} kg of pork, cancelling hamburger patty order ...");
			return null;
		}
		Console.WriteLine($"> MeatProcessingPlant: pork order successful, order state: futurePork.IsCompleted -> {futurePork.IsCompleted} ...");

		TaskCompletionSource<IReadOnlyList<HamburgerPatty>> hamburgerPattyPromise = new TaskCompletionSource<IReadOnlyList<HamburgerPatty>>();
		Task.Run(() => {
			Console.WriteLine("> MeatProcessingPlant: hiring oompa loompas from jungle ...");
			for (int i = 0; i < 10; i++) {
				ct.ThrowIfCancellationRequested();
				Thread.Sleep(100);
			}

			Console.WriteLine($"> MeatProcessingPlant: got oompa loompas, pork processing can start ... Is futurePork ready? .IsCompleted -> {futurePork.IsCompleted}");
			futurePork.ContinueWith(_ => {
				Pork pork = futurePork.Result;
				Console.WriteLine($"> MeatProcessingPlant: got {pork.Weight} kg of pork, producing hamburger patties !!! TO USE FIRST PORK !!!");

				var pattiesToProduce = (int) (pork.Weight / PorkPerHamburgerPetty);
				var producedPatties = new List<HamburgerPatty>();
				for (int i = 0; i < pattiesToProduce; i++) {
					ct.ThrowIfCancellationRequested();
					Thread.Sleep(200);
					Console.WriteLine($"> MeatProcessingPlant: 1 hamburger patty for current order produced ...");
					producedPatties.Add(new HamburgerPatty());
				}

				Console.WriteLine($"> MeatProcessingPlant: SHIPPING {producedPatties.Count} hamburger patties >> VIA PROMISE >> ...");
				hamburgerPattyPromise.SetResult(producedPatties);
			}, ct).ContinueWith(failedTask => {
				if (failedTask.IsCanceled) {
					Console.WriteLine($"> MeatProcessingPlant: processinig was cancelled, CANCELLING future hamburger patties >> VIA PROMISE >>");
					hamburgerPattyPromise.SetCanceled(ct);
				} else {
					Console.WriteLine($"> MeatProcessingPlant: processinig was cancelled, FAILING future hamburger patties >> VIA PROMISE >>");
					hamburgerPattyPromise.SetException(failedTask.Exception!);
				}
			}, TaskContinuationOptions.NotOnRanToCompletion);
		}, ct).ContinueWith(failedTask => {
			if (failedTask.IsCanceled) {
				Console.WriteLine($"> MeatProcessingPlant: oompa loompa hunting was cancelled, CANCELLING future hamburger patties >> VIA PROMISE >>");
				hamburgerPattyPromise.SetCanceled(ct);
			} else {
				Console.WriteLine($"> MeatProcessingPlant: oompa loompa hunting was cancelled, FAILING future hamburger patties >> VIA PROMISE >>");
				hamburgerPattyPromise.SetException(failedTask.Exception!);
			}
		}, TaskContinuationOptions.NotOnRanToCompletion); ;

		// Return future hamburger patties:
		return hamburgerPattyPromise.Task;
	}
}

class FastFood {
	private const int HamburgersNeeded = 7;

	public required MeatProcessingPlant HamburgerPattyProvider { get; init; }

	public void Run() {
		Console.WriteLine($"FastFood: ordering {HamburgersNeeded} hamburger patties from a MeatProcessingPlant ...");

		CancellationTokenSource cts = new CancellationTokenSource();
		// cts.Cancel();

		Task<IReadOnlyList<HamburgerPatty>>? futureHamburgerPatties = HamburgerPattyProvider.OrderHamburgerPatties(HamburgersNeeded, cts.Token);
		if (futureHamburgerPatties is not null) {
			Console.WriteLine($"FastFood: hamburger patty order successful, order state: futureHamburgerPatties.IsCompleted -> {futureHamburgerPatties.IsCompleted} ...");
			Console.WriteLine("FastFood: building new restaurant ...");
			Thread.Sleep(500);
			// Thread.Sleep(1200);
			// Thread.Sleep(2500);
			Console.WriteLine("FastFood: !FIRE! in the restaurant, cancelling hamburger patty order ...");
			cts.Cancel();

			IReadOnlyList<HamburgerPatty> hamburgers;
			try {
				hamburgers = futureHamburgerPatties.Result;
			} catch (AggregateException ex) {
				if (futureHamburgerPatties.IsCanceled) {
					Console.WriteLine($"FastFood: future hamburger patties were cancelled, TEARING DOWN the restaurant ...");
				} else {
					Console.WriteLine($"FastFood: future hamburger patties with AggregateException of:");
					foreach (var innerEx in ex.InnerExceptions) {
						Console.WriteLine($"FastFood:   '{innerEx.GetType().Name}'");
					}
				}
				return;
			}

			Console.WriteLine("FastFood: grand opening !!! TO USE FIRST HAMBURGER PATTIES !!!");
			Console.WriteLine($"FastFood: serving {hamburgers.Count} hamburgers to customers ...");
		}
	}
}

class Program {
	private static void Main(string[] args) {
		var cropFarmer = new CropFarmer(0.75f, 300);
		var pigProducer = new PigProducer { FavoriteFarmer = cropFarmer };
		var meatProcessingPlant = new MeatProcessingPlant { FavoritePigProducer = pigProducer };

		var fastFood = new FastFood { HamburgerPattyProvider = meatProcessingPlant };

		var sw = Stopwatch.StartNew();
		try {
			fastFood.Run();
		} catch (Exception ex) {
			Console.WriteLine("Running FastFood faild with exception:");
			Console.WriteLine(ex);
		}
		sw.Stop();

		Console.WriteLine($"Ellapsed time: {sw.Elapsed.TotalSeconds} days");

        Console.WriteLine("Waiting 1000 seconds ...");
        Thread.Sleep(1000);
		Console.WriteLine("Done.");
	}
}
